---------------------------------------------------------------------
-- Chess.hs
--
-- Copyright 2002,
--     Konstantin Tretjakov (kt@ut.ee)
---------------------------------------------------------------------

module Main (main) where

import Interactive
import Array
import Char

main :: IO ()
main = interact chess

-- NB!: I assume that input functions _do_not_echo_, as it is done in Hugs,
--      not as specified in the Haskell Report and done in GHC.

---------------------------------------------------------------------
-- Main types used in the program are:
--   Pt - it represents one point on a chessboard (like A1, E2, etc)
--        One special Pt value: Pt 0 0 is used to represent that
--        the user wished to quit current game.
--   Ps - represents a position in the game. Consists of 3 Pt-s:
--          the positions of White King, White Queen and Black King.
---------------------------------------------------------------------

data Pt = Pt Int Int deriving Eq
data Ps = Ps Player Pt Pt Pt


---------------------------------------------------------------------
-- Main program: chess.
--   Outputs a header message and prompts the user for initial
--   locations of game pieces. Then begins the game 
--      (playGame function)
---------------------------------------------------------------------

chess :: Interact
chess = clearScreen $
        writeStr header $
        writeStr (show (Ps player (Pt 0 0) (Pt 0 0) (Pt 0 0)) 
                 ++ "Please input initial locations of pieces:\n") $
        readPt []       (Pt (ord 'E') 1) (wKing :" White King  (default is E1): ") $ \wk ->
        readPt [wk]     (Pt (ord 'D') 1) (wQueen:" White Queen (default is D1): ") $ \wq ->
        readPt [wk, wq] (Pt (ord 'E') 8) (bKing :" Black King  (default is E8): ") $ \bk ->
        playGame (Ps machine wk wq bk) [Ps machine wk wq bk] $
        readLine "Play again? [y/N]: " $ \l ->
        writeStr (l ++ "\n") $
        if (l == "y" || l == "Y") then chess else
           writeStr "\nThanks for playing. Hope you enjoyed it.\n" $
           end


line = (replicate 80 '-')
header = ("\n" ++ line ++ "Chess Endspiel. Version 1.0.    Copyright 2002, Konstantin Tretjakov (kt@ut.ee)\n" ++ line)
bKing = '\1'
wKing = '\2'
wQueen = '\5'


---------------------------------------------------------------------
-- Show instances for Pt and Ps
---------------------------------------------------------------------

instance Show Pt where
   show (Pt c i) = (chr c):(show i)
 
instance Show Ps where
   show p@(Ps _ a b c) = concat [ showPsLine p ln | ln <- [0..18] ]
       where showPsLine pos ln
                | (ln == 0) || (ln == 18)  = "    A B C D E F G H   \n"
                | (ln == 1)  = "   ͻ  \n"
                | (ln == 17) = "   ͼ  \n"
                | (ln`mod`2 == 1) = "   Ķ  \n"
                | otherwise = " " ++ k ++  " " ++ (psLine pos row (ord 'A')) ++ " " ++ k ++ "\n"
                where row = (9 - ln`div`2) 
                      k = show row
                      psLine p@(Ps _ wk wq bk) row col
                      	| (col == (ord 'I')) = ""
                      	| (wk == Pt col row) = wKing:sep ++ pl
                      	| (wq == Pt col row) = wQueen:sep ++ pl
                      	| (bk == Pt col row) = bKing:sep ++ pl
                      	| otherwise          = ' ':sep ++ pl
                      	where sep = if col == (ord 'H') then "" else ""
                      	      pl = psLine p row (succ col)
                      	      

---------------------------------------------------------------------
-- showx - fussy "extended version" of Ps's basic show function.
-- besides showing the board also prints the moves made in the game.
-- Uses helper-functions makeHistory and showMove
-- 
-- showx.
--    Parameters:
--       pos  - board configuration to show.
--       h:hs - history of moves taken (should not be longer than 16)
--    Result:
--       outputs the screen image of a chessboard with moves history
--       on the right. The technique used is "combining pictures"
--
-- makeHistory.
--    Parameters:
--       p  - n-th position in the history
--       ps - the rest of the history
--       n  - the number of the first move to show history for
--    Result:
--       outputs a list of strings, representing moves. Moves are
--       numerated starting from n. (e.g. ["1. Kg A2 - A3+  Kg E3 - E4", "..."]
--
-- showMove.
--    Parameters:
--       p1, p2 - two game positions where p2 resulted by a move from p1
--    Result:
--       outputs a string representing the move that was made.
---------------------------------------------------------------------


showx :: Ps -> [Ps] -> String
showx pos (h:hs) = unlines (zipWith (++) (lines (show pos)) histpic)
      where histpic = ["",""] ++ makeHistory h hs 1
      
makeHistory :: Ps -> [Ps] -> Int -> [String]
makeHistory p ps n
    | length ps == 0  = [spacer ++ show n ++ ".     "] ++ repeat ""
    | length ps == 1  = [spacer ++ show n ++ ".     " ++ showMove p p1] ++ repeat ""
    | otherwise       = [spacer ++ show n ++ ".     " ++ showMove p p1 ++ "      " ++ showMove p1 p2] ++
                        makeHistory p2 (tail (tail ps)) (n+1)
                          where p1 = head ps
                                p2 = head (tail ps)
                                spacer = if n < 10 then "     " else "    "

showMove :: Ps -> Ps -> String
showMove p1@(Ps _ wk1 wq1 bk1) p2@(Ps _ wk2 wq2 bk2)
    | wk1 /= wk2  = wKing:" " ++ show wk1 ++ "-" ++ show wk2 ++ flag
    | wq1 /= wq2  = wQueen:" " ++ show wq1 ++ "-" ++ show wq2 ++ flag
    | bk1 /= bk2  = bKing:" " ++ show bk1 ++ "-" ++ show bk2 ++ flag
    where flag = if checkmate p2 then " X" 
                 else if onFire p2 bk2 then " +"
                 else "  "


---------------------------------------------------------------------
-- Interactive combinator readPt: inputs a Pt from the user
--    Parameters: 
--       usedPts - a list of points that the user may not enter
--       defaultPt - a point to be chosen if the user enters nothing
--       prompt - a prompt string
---------------------------------------------------------------------

readPt :: [Pt] -> Pt -> String -> (Pt -> Interact) -> Interact
readPt usedPts defaultPt prompt prog
        = readLine prompt $ \l ->
          let t = filter (/=' ') l
              c = toUpper (head t)
              r = head (tail t)
	      pt = Pt (ord c) (read [r])
              redo = readPt usedPts defaultPt  "  Invalid input, please enter a location (e.g. A1, C2, ..): " prog
              redo2 = readPt usedPts defaultPt "  This location is already used, choose another one: " prog
           in if t == "" then 
          	writeStr (shows defaultPt "\n") $
          	if defaultPt`elem`usedPts then redo2 else (prog defaultPt)
             else 
               writeStr (l ++ "\n") $
               if (length t /= 2) then redo
               else if (c > 'H' || c < 'A' || r > '8' || r < '1') then redo
               else if (pt`elem`usedPts) then redo2
               else (prog pt)


---------------------------------------------------------------------
-- Function playGame. Interchanges moves between players.
-- Checks for checkmate/stalemate condition and quits game if so.
--    Parameters:
--       pos - current board configuration
--       hist - game history: a list of positions
--              in order they happened in the game.
---------------------------------------------------------------------

playGame :: Ps -> [Ps] -> Interact -> Interact
playGame pos@(Ps curPlayer wk wq bk) hist prog
        = clearScreen $
          writeStr ("\n" ++ showx pos hist ++ "\n") $
          if (curPlayer == machine && onFire pos bk) then
             writeStr ("                           Illegal position!\n\n") prog
          else if checkmate pos then 
             writeStr ("                           Checkmate! White wins!\n\n") prog
          else if stalemate pos then
             writeStr ("                           Stalemate!\n\n") prog
          else if (curPlayer == player) then
             readMove ("Please enter Q/W/E/A/D/Z/X/C to move the black king in the\n" ++
                       " corresponding direction. Enter T to abort game.\n") pos $ \newpos@(Ps _ wk wq bk) ->
             if bk == (Pt 0 0) then 
                writeStr ("                           Game aborted!\n\n") prog
             else playGame newpos (hist ++ [newpos]) prog
          else 
             writeStr "Please wait... \n" $
             playGame newmove (hist ++ [newmove]) prog
             where newmove = makeAmove pos  -- NB! makeMove 1 pos works incorrectly here.


---------------------------------------------------------------------
-- Interactive combinator readMove. Prompts the user to enter his
-- move. Assumes the situation on the board is legal. Assures,
-- that the player makes a valid move. Returns the new position.
--    Parameters:
--       prompt - a string to prompt user for his move.
--       pos - current board configuration
--       prog - (Ps->Interact), that gets the new board 
--              configuration as input.
---------------------------------------------------------------------

readMove :: String -> Ps -> (Ps -> Interact) -> Interact
readMove prompt pos@(Ps pl wk wq (Pt col row)) prog
        = writeStr prompt $
          readChar $ \c ->
          let uc = toUpper c
              newpt = case uc of 
                        'A' -> Pt (pred col) row
                        'D' -> Pt (succ col) row
                        'W' -> Pt col (succ row)
                        'X' -> Pt col (pred row)
                        'Q' -> Pt (pred col) (succ row)
                        'E' -> Pt (succ col) (succ row)
                        'C' -> Pt (succ col) (pred row)
                        'Z' -> Pt (pred col) (pred row)
                        'T' -> Pt 0 0      -- Means user wants to quit
                	_   -> Pt (-1) 0   -- This line is not needed actually
              newps = Ps (opponent pl) wk wq newpt
                                                       
           in if not (uc`elem`"QWEADZXCT") then readMove "" pos prog
              else if illegalPt newpt then readMove "Cannot move there, you are at the board boundary!\n" pos prog
              else if onFire pos newpt then readMove "Cannot move there, the location is under fire!\n" pos prog
              else prog newps


---------------------------------------------------------------------
-- Chess and Game Logic functions.
---------------------------------------------------------------------

---------------------------------------------------------------------
-- Function illegalPt.
-- Returns True if the given Pt is a legal board location.
-- Pt 0 0 is considered a special legal Pt value. 
-- It means that user wanted to abort current game.
---------------------------------------------------------------------

illegalPt :: Pt -> Bool
illegalPt p@(Pt col row) = not (col >= (ord 'A') && col <= (ord 'H') && row >=1 && row <= 8) 
                                            && not (p == Pt 0 0)

                                         
---------------------------------------------------------------------
-- Functions kingMoves, queenMoves.
-- Return a list of Pt's where a king/queen may move from a given 
-- location assuming there are no other pieces on the board.
---------------------------------------------------------------------                                        

kingMoves, queenMoves :: Pt -> [Pt]
kingMoves (Pt col row) = [(Pt (col + i) (row + j)) | 
                             i <- [-1..1], j <- [-1..1],
                             (i/=0 || j/=0),
                             not (illegalPt (Pt (col + i) (row + j)))]  

queenMoves (Pt col row) = [(Pt (col + i) (row + j)) |
                             i <- [-7..7], j <- [-7..7],
                             (i/=0 || j/=0),
                             (i == 0 || j == 0 || i == j || i == -j),
                             not (illegalPt (Pt (col + i) (row + j)))]


---------------------------------------------------------------------
-- Function notObstacle a b c.
-- Returns True, if Pt a is not an obstacle on the way from b to c
-- (assuming that b and c are on one horizontal/vertical/diagonal)
---------------------------------------------------------------------

notObstacle :: Pt -> Pt -> Pt -> Bool
notObstacle (Pt xa ya) (Pt x1 y1) (Pt x2 y2) 
   = let dx1 = xa - x1
         dx2 = xa - x2
         dy1 = ya - y1
         dy2 = ya - y2
     in not (dx1*dy2 - dx2*dy1 == 0 && 
            (x1 <= xa && xa <= x2 || x1 >= xa && xa >= x2) &&
            (y1 <= ya && ya <= y2 || y1 >= ya && ya >= y2))


---------------------------------------------------------------------
-- onFire - returns True, if location Pt is under white fire
--    Parameters:
--       pos - board configuration
--       pt - location to test 
---------------------------------------------------------------------

onFire :: Ps -> Pt -> Bool
onFire pos@(Ps _ wk@(Pt wkx wky) wq@(Pt wqx wqy) _) pt@(Pt x y) 
   = (abs (x - wkx) == 1 && abs (y - wky) <= 1) ||
     (abs (x - wkx) <= 1 && abs (y - wky) == 1) ||
     (((x == wqx) || (y == wqy) || abs (x - wqx) == abs (y - wqy)) && (wq /= pt) && notObstacle wk wq pt)
 

---------------------------------------------------------------------
-- Functions checkmate and stalemate: check for relevant states
-- of the board. (Obvious that checkmated/stalemated can be only
-- the black king). 
---------------------------------------------------------------------

checkmate :: Ps -> Bool
checkmate ps@(Ps pl wk wq bk) = foldr (&&) True [ onFire ps pt | pt <- bk : kingMoves bk ] 

stalemate :: Ps -> Bool
stalemate ps@(Ps pl wk wq bk) = (pl == player) && not (onFire ps bk) && 
                                   foldr (&&) True [ onFire ps pt | pt <- kingMoves bk ]


---------------------------------------------------------------------
-- Position instance for Ps.
---------------------------------------------------------------------

instance Position Ps where

   ---------------------------------------------------------------------
   -- Function dynamic.
   -- Should return True, if the position is too "dynamic" to be
   -- statically evaluated. In our case here every position could
   -- be evaluated statically. So dynamic is always false.
   ---------------------------------------------------------------------

   dynamic pos = False


   ---------------------------------------------------------------------
   -- Function win.
   -- Should return True, if the position is winning for a given 
   -- player. In our case this function is actually not needed at all,
   -- but it is to be defined for completeness. Winning position for the
   -- white is a checkmate and for the black - stalemate.
   -- The first argument of the function (player) is ignored, because
   -- the player information is already contained in the position.
   ---------------------------------------------------------------------

   win _ pos@(Ps pl a b c) = if pl == machine then checkmate pos else stalemate pos


   ---------------------------------------------------------------------
   -- Function moves.
   -- Returns a list of positions, that could be obtained from a given
   -- legal position by making a legal move
   ---------------------------------------------------------------------
  
   moves pos@(Ps pl wk wq bk@(Pt bkc bkr))
      | pl == player  =  [Ps machine wk wq nbk | 
                                 nbk <- kingMoves bk,
                                 not (onFire pos nbk)]  
      | pl == machine =  [Ps player (Pt nwkc nwkr) wq bk  |
                                 Pt nwkc nwkr <- kingMoves wk,
                                 (Pt nwkc nwkr) /= wq,
                                 max (abs (nwkc - bkc)) (abs (nwkr - bkr)) > 1] ++
                         [Ps player wk nwq bk  |
                                 nwq <- queenMoves wq,
                                 nwq /= wk,
                                 notObstacle wk nwq wq]

           
   ---------------------------------------------------------------------
   -- Function static.
   -- The "AI" of the game - statically evaluates the position
   -- from the point of view of the white player (the computer, that is).
   -- So the checkmate position has the highest mark,
   -- stalemate - the lowest.
   --
   -- Position evaluation scheme.
   -- The main stress is put on the "black king freedom", i.e. the 
   -- number of locations reachable by the black king. It is clear that 
   -- one of the winning strategies for the white is to reduce this
   -- freedom until it reaches 0. The freedom may nearly always be 
   -- be reduced only by the actions of the white queen, but in some
   -- cases the white king should be near the queen. So the distance
   -- between the white king and queen is also considered, as a secondary
   -- factor. Having the opportunity to look forward 2 moves and ignoring
   -- the stalemate condition, such a scheme would be sufficient.
   -- The two problems that arise are 
   --   1) Inability to search 2 moves deep
   --   2) Avoiding the stalemate.
   -- Therefore some hacks were included in the scheme, mostly
   -- found by trial and error, that are supposed to solve the problems.
   --   1) A stalemate condition, being here equal to white's loss, is
   --      always marked as 0 so there is no possibility for the white
   --      to produce a stalemate.
   --   2) There is also a non-stalemate condition, where the black
   --      king is under fire, and his freedom degree is 1.
   --      It was found, that this condition often leads to "dead loops"
   --      So it is also marked as bad.
   --   3) And in general, whenever the freedom of black king reaches 2
   --      there is a strong possibility of a "dead loop", that seemes
   --      to have been solved by motivating the white king move closer to 
   --      the black one. However, white king is "not advised" to move to 
   --      the same border, where the trapped black king is. It should
   --      try to "push" the black king near the boundary. 
   --      This is done by the "kingsOfBorder" mark in the formula.
   -- It seems pretty hard to prove, that the scheme is always successful
   -- and solves the task as it should (in 20 moves), but numerous tests
   -- show that it does.  
   ---------------------------------------------------------------------


   static ps@(Ps _ wk@(Pt wkc wkr) wq@(Pt wqc wqr) bk@(Pt bkc bkr))
      | (abs (wqc - bkc) <= 1 && abs (wqr - bkr) <= 1) && not (onFire ps wq) = 0
      | stalemate ps = 0
      | freeBk == 0  = 1000
      | freeBk == 1  = 1
      | freeBk <= 2  = 1000 - 10 - (abs (wkc - bkc) + abs (wkr - bkr)) - kingsOnBorder
      | otherwise    = 1000 - 10 * freeBk - dist
                    where blank = array bnds [(i, False) | i <- range bnds ]
                          bnds  = ((ord 'A', 1), (ord 'H', 8))
                          mvs = if onFire ps bk then kingMoves bk else (bk : kingMoves bk)
                          freeBk = blackFreedom ps mvs (blank//[((a,b),True) | Pt a b <- mvs]) 0
                          dist = max (abs (wqc - wkc)) (abs (wqr - wkr))
                          kingsOnBorder = if ((wkc == ord 'A' || wkc == ord 'H') && wkc == bkc)
                                           ||((wkr == 1 || wkr == 8) && wkr == bkr) then 5 else 0
           

---------------------------------------------------------------------
-- Function blackFreedom.
-- Finds the amount of locations reachable by the black king.
-- Uses a kind of a simple dfs. Arrays are used to improve 
-- efficiency.
--    Parameters:
--       ps    - configuration of the board (the locations of 
--               the white pieces are of importance).
--       moves - a list of locations to be tested, where the black 
--               king could probably move.
--       board - an ('A'..'H' x 1..8) array, where locations marked
--               as True correspond to those, that are already taken
--               in consideration.
--       n     - accumulator. It's value corresponds to the number
--               of locations reachable for the black king among
--               those, that have already been examined. 
---------------------------------------------------------------------
                                
blackFreedom :: Ps -> [Pt] -> Array (Int,Int) Bool -> Int -> Int

blackFreedom ps [] board n = n
blackFreedom ps@(Ps _ wk wq bk) moves@(m@(Pt a b):ms) board n
   | (onFire ps m)= blackFreedom ps ms board n
   | otherwise    = blackFreedom ps (newlocs ++ ms) (board//[((a,b),True) | Pt a b <- newlocs]) (n+1)
                    where newlocs = [Pt a b | (Pt a b) <- kingMoves m, board!(a,b) == False]


---------------------------------------------------------------------
-- makeAmove and findBest - functions, that find the best move for
-- the white player. They actually perform the functions of 
-- the Minimax module and here's why they are here.
--
-- Basically the initial idea was to use the Minimax module, with
-- the help of which I could use a much simpler position evaluation
-- scheme without the heuristical hacks. The only thing I needed
-- would be the possibility of looking 2 white moves forward.
-- Common sense tells, that such evaluation requires an order of
-- 30*8*30 = 4600 evaluations which is not a number, is it?
-- But, as it occured, even this evaluation takes unacceptably much
-- time and the only acceptable search depth is therefore 1 move.
-- These circumstances forced the complication of the evaluation
-- scheme (and my inability now to assert its correctness).
--
-- Evaluation of a game tree 1 move depth doesn't really require
-- all the gametree apparatus, so makeAmove and findBest just
-- improve the performance.
---------------------------------------------------------------------

makeAmove :: Ps -> Ps
makeAmove pos = findBest (moves pos) pos 0

findBest :: [Ps] -> Ps -> Int -> Ps
findBest [] bestpos bestmark = bestpos
findBest (p:ps) bestpos bestmark 
    = if newbest > bestmark then (findBest ps p newbest) else (findBest ps bestpos bestmark)
         where newbest = static p


---------------------------------------------------------------------
-- Data that was formerly legacy of module Minimax.
---------------------------------------------------------------------

class Position pos where
  moves   :: pos -> [pos]
  static  :: pos -> Int
  dynamic :: pos -> Bool
  win     :: Player -> pos -> Bool

type Player = Bool

machine, player :: Player
machine = True
player  = False

opponent :: Player -> Player
opponent = not